/*
 * CONTENT BASED SCHEDULING / WEIGHTED CONTENT BASED SCHEDULING
 */

#define __KERNEL_SYSCALLS__             /*  for waitpid */

#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/vmalloc.h>
#include <linux/net.h>
#include <linux/skbuff.h>
#include <linux/unistd.h>
#include <linux/wait.h>
#include <asm/unistd.h>

#include <net/sock.h>
#include <net/tcp.h>

#include "tcpslb.h"
#include "cbs.h"
#include "common.h"

static int tcpslb_accept_cons_cbs(void *service);
static int tcpslb_cross_conn(void *data);
static int tcpslb_serv_other_side(void *data);
static struct tcpslb_realservers *tcpslb_next_cbs(struct tcpslb_service *serv);
static int tcpslb_cbs_counter(void *service);
static struct tcpslb_realservers *tcpslb_next_wcbs(struct tcpslb_service *serv);
static int tcpslb_wcbs_counter(void *service);
extern __u32 in_aton(const char *str);
static int errno;

static int __init tcpslb_cbs_init(void)
{
	inter_module_register("tcpslb_accept_cons_cbs", THIS_MODULE, tcpslb_accept_cons_cbs);
		      return 0;
}

static void __exit tcpslb_cbs_cleanup(void)
{
  	inter_module_unregister("tcpslb_accept_cons_cbs");
}
 
/*
 *  Accept connections for cbs & wcbs
 */
static int tcpslb_accept_cons_cbs(void *service)
{
        struct socket *sock,*sock1;
	int ret;
	struct tcpslb_cons *conn;
	struct tcpslb_service *serv=(struct tcpslb_service *)service;

	//create a listening socket
        if(tcpslb_create_listening_sock(&sock,serv)!=0)	{
		return -1;
	}
	/*
	 * Here we create a thread which will maintain the counter for the
	 * content which passed through server.. The purpose here is to prevent
	 * the integer maintainig the counter from exceeding the integer limit.
	 */
	if(serv->scheduling==SCHEDULE_CBS)
		kernel_thread(&tcpslb_cbs_counter,service,0);
	else if(serv->scheduling==SCHEDULE_WCBS)
		kernel_thread(&tcpslb_wcbs_counter,service,0);
		
	//tell the world that the daemon is up
	down_interruptible(&serv->sem); //check ret values if err ret -ERESTARTSYS

	
	while(1){  //infinite loop to keep accepting conns

		//start acceptin clients
	  if((ret=tcpslb_accept(sock,&sock1,&serv->tcpslb_cont))<0){
		  perror("accept");
	          break; //got -2  i.e. cleanup has been triggerd
     	  }

	  conn=(struct tcpslb_cons *) vmalloc(sizeof(struct tcpslb_cons));

	  //fill up the structure... since we have a client waiting for service
	  conn->srcSock  = sock1; //sock1 must be de allocated
	  conn->service  = serv;
	  if(serv->scheduling==SCHEDULE_CBS)
	        conn->realServ = tcpslb_next_cbs(serv);
	  else if(serv->scheduling==SCHEDULE_WCBS)
		conn->realServ = tcpslb_next_wcbs(serv);
			
	  //we keep threading for each obtained client
          kernel_thread(&tcpslb_cross_conn,(void *)conn,0);

	}
	sock_release(sock); //close the listening socket
	up(&serv->sem);//tell cleanup the daemon is returning
	return 0;
}



/*
 * this Function will open the socket to the real server. and serve one side
 * side of the connection.
 */
static int tcpslb_cross_conn(void *data)
{
	
	char *msg;
	int len;
	struct socket *sock;
	struct tcpslb_cons *conn=(struct tcpslb_cons *)data; 
	//this must be deallocated at the end of this function
	static struct sockaddr_in mysock;
	struct tcpslb_sockpair_cbs *sockpair;

	struct tcpslb_realservers *server;
	struct tcpslb_sched_cbs *cbs_sched;
	struct tcpslb_sched_wcbs *wcbs_sched;

	unsigned long int *content=NULL,i;

	//determine the number of the scheduled server
	for(i=0,server=conn->service->realServ;
			i<conn->service->noRealServ && server!=conn->realServ;
			i++,server=server->next);

	//if we did not find the scheduled server then it must have just
	//gone down
	if(i==conn->service->noRealServ){
		perror("Some Real Svr went down??");
		//we may want to reschedule here
		return 0;
	}

	//setting a pointer content which will contain the address of 
	//the counter
	if(conn->service->scheduling==SCHEDULE_CBS){
		cbs_sched=(struct tcpslb_sched_cbs *)conn->service->data;
		content=&cbs_sched->content[i];
	}
	else if(conn->service->scheduling==SCHEDULE_WCBS){
		wcbs_sched=(struct tcpslb_sched_wcbs *)conn->service->data;
		content=&wcbs_sched->content[i];
	}
        
	//a stream socket to real server
	if (sock_create(AF_INET, SOCK_STREAM, IPPROTO_TCP, &sock) < 0) {
           perror("Error during creation of socket; terminating\n");
           return 0;
        }

	//setting in the values of the server as we know it
	memset(&mysock, 0, sizeof(mysock));
        mysock.sin_family = AF_INET;
        mysock.sin_port = htons(conn->realServ->port);
	//in_aton is giving a warning.. but working fine..
        mysock.sin_addr.s_addr = 
		in_aton(conn->realServ->addr);//addr
#ifdef DEBUG_MESSAGE			
	perror(conn->realServ->addr);
#endif

	//connect to the real server
       if(sock->ops->connect(sock,(struct sockaddr *)&mysock,
			       sizeof(mysock),0)<0){
		printk("Hit Error!!");
	        sock_release(sock);
 	        sock_release(conn->srcSock);
	        vfree(conn); 
		return 0;
	}

       //creating the sockpair.. to contain the pair of sockets to be spliced
	sockpair = (struct tcpslb_sockpair_cbs *)
		vmalloc(sizeof(struct tcpslb_sockpair_cbs));
	//must de alloc this
	sockpair->srcSock=conn->srcSock;
	sockpair->destSock=sock;
	sockpair->cont=1;//this is to continue the execution of both the threads
	sockpair->content=content;
	
	if(conn->service->scheduling==SCHEDULE_CBS){
		sockpair->cbs_sched=(struct tcpslb_sched_cbs *)
			conn->service->data;
		sockpair->wcbs_sched=NULL;
	}
	else if(conn->service->scheduling==SCHEDULE_WCBS){
		sockpair->wcbs_sched=(struct tcpslb_sched_wcbs *)
			conn->service->data;
		sockpair->cbs_sched=NULL;
	}

        //initializing the semaphore... to know status of other side of service
	sema_init(&(sockpair->sem),1);
        //we also need a spin lock to control value of sockpair->cont
	spin_lock_init(&(sockpair->lock));

	//thread again so that we can read @ both sock
        kernel_thread(&tcpslb_serv_other_side,(void *)sockpair,0);
	
	msg=(char*) get_free_page(GFP_KERNEL);

        //we continule till this is 1
	while(sockpair->cont){

		//while the queue is empty and we can continue
		while(skb_queue_empty(&(sockpair->destSock->sk->receive_queue))
				&& sockpair->cont) {

		  /* if we are the socket is not in state TCP_ESTABLSIHED
		   * we must quit by setting sockpair->cont to 0
		   * also if the state is TCP_CLOSE_WAIT and the receive buffer
		   * is empty we must set sockpair->cont to 0 since server has 
		   * closed communication and we are in close state...
		   * 
		   * These statements have some bug!! ... the procedure is 
		   * working fine with telnet (to any port)... but clients like
		   * browsers hold connection with server for a long time
		   */ 
//	              printk("d: %d q: %d --",sockpair->destSock->sk->state,
//		      skb_queue_empty(&sockpair->destSock->sk->receive_queue));
     		  if(sockpair->destSock->sk->state != TCP_ESTABLISHED && 
		    !(sockpair->destSock->sk->state==TCP_CLOSE_WAIT && 
		    !skb_queue_empty(&sockpair->destSock->sk->receive_queue))){

			  //set the value to 0 and keep it atomic least it be 
			  //modified halfway by some party..
			  spin_lock(&(sockpair->lock));
		  	  sockpair->cont=0; 
			  spin_unlock(&(sockpair->lock));
	  		  break;
		   }
		  /* If we reach here, the queue is empty... 
		   * so we go on interruptible sleep
		   */
		   interruptible_sleep_on_timeout(&sock->wait, HZ);
//	              printk("dw: %d q: %d --",sockpair->destSock->sk->state,
//		      skb_queue_empty(&sockpair->destSock->sk->receive_queue));
		}
		
		if(sockpair->cont==0) break;

		//receive data... on queue
		if((len=tcpslb_receive(sockpair->destSock, msg, PAGE_SIZE))<=0){
#ifdef DEBUG_MESSAGE			
		     	perror("got out @ recv1");
			printk("-%d-",len);
#endif
	     		break;
		}

		//transmit it
		if(tcpslb_send(sockpair->srcSock,msg,len)<=0){
#ifdef DEBUG_MESSAGE			
		     	perror("got out @ send1");
#endif
			break;
		}
		
		//now that we transmitted the data. we must increment the 
		//counter.. that is what we do next
		if(sockpair->cbs_sched==NULL)
			down_interruptible(&sockpair->wcbs_sched->sem);
		else
			down_interruptible(&sockpair->cbs_sched->sem);
		
		if(*(sockpair->content)<4000000000)
		*(sockpair->content) += len;
		
		if(sockpair->cbs_sched==NULL)
			up(&sockpair->wcbs_sched->sem);
		else
			up(&sockpair->cbs_sched->sem);

		
//	              printk("ds: %d q: %d --",sockpair->destSock->sk->state,
//		      skb_queue_empty(&sockpair->destSock->sk->receive_queue));
	}
	
	//if we get out of while w/o setting sockpair->cont to zero
	//a send / receive error has occured... closing other side
	spin_lock(&(sockpair->lock));
	if(sockpair->cont==1)  sockpair->cont=0;
	spin_unlock(&(sockpair->lock));

	//this side of service is closed... the other side will take some 
	//timeout time to close... to remove this delay we wake up the other
	//side
	wake_up_interruptible(&sockpair->srcSock->wait);

        //checking of other side closed
	down_interruptible(&(sockpair->sem));

	//clear all the resource associated with this connection
	sock_release(sock);
	sock_release(conn->srcSock);
	vfree(conn); 
	vfree(sockpair); //deallocetd here!!!!

	free_page((unsigned long)msg);
	//release the zombie created by service on other side
	 waitpid(-1, NULL, __WNOTHREAD);
	return 0;
	
}

/*
 * This is essential the same funcion as above with the connection to real 
 * server part removed. Only differnce is it acts on opposite sides of the 
 * socket serving the other side..
 */
static int tcpslb_serv_other_side(void *data)
{
	struct tcpslb_sockpair_cbs *sockpair = 
		(struct tcpslb_sockpair_cbs *)data;
	char   *msg;
	int len;

	//semaphore locked till this thread lives
	down_interruptible(&(sockpair->sem));
	msg=(char*) get_free_page(GFP_KERNEL);
	
	while(sockpair->cont){
		
		while (skb_queue_empty(&(sockpair->srcSock->sk->receive_queue))
				&& sockpair->cont) {
 //		  printk("s: %d q: %d --",sockpair->srcSock->sk->state,
// 		  skb_queue_empty(&sockpair->srcSock->sk->receive_queue));
		
		  if(sockpair->srcSock->sk->state != TCP_ESTABLISHED && 
		   !(sockpair->srcSock->sk->state==TCP_CLOSE_WAIT && 
		   !skb_queue_empty(&sockpair->srcSock->sk->receive_queue))){
			  spin_lock(&(sockpair->lock));
			  sockpair->cont=0; 
			  spin_unlock(&(sockpair->lock));
			  break;
		  }
		  interruptible_sleep_on_timeout(&sockpair->srcSock->wait, HZ);
 //		  printk("sw: %d q: %d --",sockpair->srcSock->sk->state,
// 		  skb_queue_empty(&sockpair->srcSock->sk->receive_queue));
		}
		
		if(sockpair->cont==0) break;
	
	     	/* 
		 * Note that this is a non blocking receive..since we have to 
		 * look @ both the received value and also see if can continue
		 * the daemon...
		 */
		if((len=tcpslb_receive(sockpair->srcSock, msg,PAGE_SIZE))<=0){
#ifdef DEBUG_MESSAGE			
	     		perror("got out @ recv2");
#endif
			break;
	     	}
	   
		if((len=tcpslb_send(sockpair->destSock, msg,len))<=0){
#ifdef DEBUG_MESSAGE			
		     	perror("got out @ send2");
#endif
			break;
		}
		if(sockpair->cbs_sched==NULL)
			down_interruptible(&sockpair->wcbs_sched->sem);
		else
			down_interruptible(&sockpair->cbs_sched->sem);
		if(*(sockpair->content)<4000000000)
		*(sockpair->content)+=len;
		if(sockpair->cbs_sched==NULL)
			up(&sockpair->wcbs_sched->sem);
		else
			up(&sockpair->cbs_sched->sem);
 //		  printk("ss: %d q: %d --",sockpair->srcSock->sk->state,
// 		  skb_queue_empty(&sockpair->srcSock->sk->receive_queue));
	
	}
	
	spin_lock(&(sockpair->lock));
	if(sockpair->cont==1)  sockpair->cont=0;
	spin_unlock(&(sockpair->lock));

	wake_up_interruptible(&sockpair->destSock->wait);
	free_page((unsigned long)msg);

	up(&(sockpair->sem)); 
	//thread terminationg so we clear the semaphore.. telling the 1st side
	//that it may clear up the resources.

	return 0;
}

/*
 * This procedure implements the content based scheduling.. the next server
 * is determined by ->          min(content)
 */
static struct tcpslb_realservers *tcpslb_next_cbs(struct tcpslb_service *serv)
{
	struct tcpslb_sched_cbs *sched=(struct tcpslb_sched_cbs *)serv->data;
	struct tcpslb_realservers *server;
	int i,svr=0,min=99999;
 	   
	for(i=0;i<serv->noRealServ;i++){
     		if(sched->content[i]<min){
			min=sched->content[i];
			svr=i;
		}
	}

	for(i=0,server=serv->realServ;i<svr;i++,server=server->next);
	return server;
			
}

/*
 * This function prevents the counter from exceeding the integer limit..
 * here we find the minimum content value and deduct it from the rest.
 */
static int tcpslb_cbs_counter(void *service)
{
	DECLARE_WAIT_QUEUE_HEAD(wait);
	struct tcpslb_service *serv=(struct tcpslb_service *)service;
	struct tcpslb_sched_cbs *sched=(struct tcpslb_sched_cbs *)serv->data;
	int i,min;
	down_interruptible(&sched->sem1);
	
	//find the minumum
	while(serv->tcpslb_cont){
           min=99999;
 	   for(i=0;i<serv->noRealServ;i++){
		   if(sched->content[i]<min){
			   min=sched->content[i];
		   }
	   }
	   
	   //if minimum is not zero deduct it from rest of the values
	   if(min!=0){
		   down_interruptible(&sched->sem);
		   for(i=0;i<serv->noRealServ;i++){
			   sched->content[i]-=min;
		   }
		   up(&sched->sem);
	   }
	   //printk("-%ld %ld-",sched->content[0],sched->content[1]);

	   //sleep for timeout period
           interruptible_sleep_on_timeout(&wait,HZ);
	}

	//notify cleanup that this thread is done
	up(&sched->sem1);
	
	return 0;
}

/*
 * This procedure implements the weighted content based scheduling.. 
 * the next server is determined by ->
 * min(content[i]/weight[i] ) for all i
 */
static struct tcpslb_realservers *tcpslb_next_wcbs(struct tcpslb_service *serv)
{
	struct tcpslb_sched_wcbs *sched=(struct tcpslb_sched_wcbs *)serv->data;
	struct tcpslb_realservers *server;
	int i,svr=0,min=99999;
 	   
	for(i=0;i<serv->noRealServ;i++){
     		if(sched->content[i]/sched->weight[i]<min){
			min=sched->content[i]/sched->weight[i];
			svr=i;
		}
	}

	for(i=0,server=serv->realServ;i<svr;i++,server=server->next);
	return server;
			
}

/*
 * This function prevents the counter from exceeding the integer limit..
 * consider the derivation..
 *   content[i]_new / weight[i] =  
 *        content[i]_old/weight[i] - content[min] / weight[min]
 *        where min=is the sever having min content/weight value
 *              content[i]_new is to be determined
 *    so we have,
 *     content[i] = (content[i]/weight[i]-content[min]/weight[min])
 *                   * weight[i]
 *     This formula is what exactly what the procedure implements
 */
static int tcpslb_wcbs_counter(void *service)
{
	DECLARE_WAIT_QUEUE_HEAD(wait);
	struct tcpslb_service *serv=(struct tcpslb_service *)service;
	struct tcpslb_sched_wcbs *sched=(struct tcpslb_sched_wcbs *)serv->data;
	int i,min,svr;
//	sched->sleepq=&wait;
	down_interruptible(&sched->sem1);
	while(serv->tcpslb_cont){
           min=99999;
	   svr=0;
 	   for(i=0;i<serv->noRealServ;i++){
		   if(sched->content[i]/sched->weight[i]<min){
			   min=sched->content[i]/sched->weight[i];
			   svr=i;
		   }
	   }
	   
	   if(min!=0){
	    down_interruptible(&sched->sem);
	    for(i=0;i<serv->noRealServ;i++){
	      sched->content[i]=( (sched->content[i] /sched->weight[i]) 
		      -min) * sched->weight[i];
		      if(sched->content[i]<0) sched->content[i]=0;
	   }
		   up(&sched->sem);
	  }
//	   printk("-%d %d-",sched->content[0],sched->content[1]);
           interruptible_sleep_on_timeout(&wait,HZ);
	}
	up(&sched->sem1);
	
	return 0;
}

module_init(tcpslb_cbs_init);
module_exit(tcpslb_cbs_cleanup);
MODULE_LICENSE("GPL");

